==> 0.1.32.zip: oidc-fastapi-test/src/oidc_test/auth_utils.py <==
from typing import Union, Annotated, Tuple
from functools import wraps
import logging

from fastapi import HTTPException, Request, Depends, status
from fastapi.security import OAuth2PasswordBearer
from authlib.oauth2.rfc6749 import OAuth2Token
from authlib.integrations.starlette_client import OAuth, OAuthError, StarletteOAuth2App
from jwt import ExpiredSignatureError, InvalidKeyError, decode
from httpx import AsyncClient

# from authlib.oauth1.auth import OAuthToken
# from authlib.oauth2.auth import OAuth2Token

from .models import User
from .database import db, UserNotInDB
from .settings import settings, OIDCProvider

logger = logging.getLogger(__name__)

oidc_providers_settings: dict[str, OIDCProvider] = dict([(provider.id, provider) for provider in settings.oidc.providers])

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")

def fetch_token(name, request):
    breakpoint()
    ...
    # if name in oidc_providers:
    #    model = OAuth2Token
    # else:
    #    model = OAuthToken

    # token = model.find(name=name, user=request.user)
    # return token.to_token()


def update_token(*args, **kwargs):
    breakpoint()
    ...


authlib_oauth = OAuth(cache=None, fetch_token=fetch_token, update_token=update_token)


def init_providers():
# Add oidc providers to authlib from the settings
    for id, provider in oidc_providers_settings.items():
        authlib_oauth.register(
            name=id,
            server_metadata_url=provider.openid_configuration,
            client_kwargs={
                "scope": "openid email offline_access profile",
            },
            client_id=provider.client_id,
            client_secret=provider.client_secret,
            api_base_url=provider.url,
            # For PKCE (not implemented yet):
            # code_challenge_method="S256",
            # fetch_token=fetch_token,
            # update_token=update_token,
            # client_id="some-client-id", # if enabled, authlib will also check that the access token belongs to this client id (audience)
        )

init_providers()

async def get_providers_info():
    # Get the public key:
    async with AsyncClient() as client:
        for provider_settings in oidc_providers_settings.values():
            if provider_settings.info_url:
                provider_info = await client.get(provider_settings.url)
                provider_settings.info = provider_info.json()


def get_oidc_provider_or_none(request: Request) -> StarletteOAuth2App | None:
    """Return the oidc_provider from a request object, from the session.
    It can be used in Depends()"""
    if (oidc_provider_id := request.session.get("oidc_provider_id")) is None:
        return
    try:
        return getattr(authlib_oauth, str(oidc_provider_id))
    except AttributeError:
        return


def get_oidc_provider(request: Request) -> StarletteOAuth2App:
    if (oidc_provider := get_oidc_provider_or_none(request)) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No such provider")
    else:
        return oidc_provider


async def get_current_user(request: Request) -> User:
    """Get the current user from a request object.
    Also validates the token expiration time.
    ... TODO: complete about refresh token
    """
    if (user_sub := request.session.get("user_sub")) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED)
    if (token := await db.get_token(request.session["token"])) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Token unknown")
    user = await db.get_user(user_sub)
    ## Check if the token is expired
    if token.is_expired():
        oidc_provider = get_oidc_provider(request=request)
        ## Ask a new refresh token from the provider
        logger.info(f"Token expired for user {user.name}")
        try:
            userinfo = await oidc_provider.fetch_access_token(
                refresh_token=token.get("refresh_token")
            )
        except OAuthError as err:
            logger.exception(err)
            # raise HTTPException(
            #    status.HTTP_401_UNAUTHORIZED, "Token expired, cannot refresh"
            # )

    return user


async def get_token(request: Request) -> OAuth2Token:
    """Return the token from a request object, from the session.
    It can be used in Depends()"""
    if (token := await db.get_token(request.session.get("token"))) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No token")
    return token


async def get_current_user_or_none(request: Request) -> User | None:
    """Return the user from a request object, from the session.
    It can be used in Depends()"""
    try:
        return await get_current_user(request)
    except HTTPException:
        return None


def hasrole(required_roles: Union[str, list[str]] = []):
    """Decorator for RBAC permissions"""
    required_roles_set: set[str]
    if isinstance(required_roles, str):
        required_roles_set = set([required_roles])
    else:
        required_roles_set = set(required_roles)

    def decorator(func):
        @wraps(func)
        async def wrapper(request=None, *args, **kwargs):
            if request is None:
                raise HTTPException(
                    500,
                    "Functions decorated with hasrole must have a request:Request argument",
                )
            user: User = await get_current_user(request)
            if not any(required_roles_set.intersection(user.roles_as_set)):
                raise HTTPException(status.HTTP_401_UNAUTHORIZED)
            return await func(request, *args, **kwargs)

        return wrapper

    return decorator


def get_token_info(token: dict) -> dict:
    token_info = dict()
    for key in token:
        if key != "userinfo":
            token_info[key] = token[key]
    return token_info


async def get_user_from_token(
    token: Annotated[str, Depends(oauth2_scheme)],
    request: Request,
) -> User:
    if (auth_provider_id := request.headers.get("auth_provider")) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Request headers must have a 'auth_provider' field")
    if (auth_provider_settings := oidc_providers_settings.get(auth_provider_id)) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, f"Unknown auth provider '{auth_provider_id}'")
    oidc_provider: StarletteOAuth2App = getattr(authlib_oauth, auth_provider_id)
    await oidc_provider.load_server_metadata()
    if (key := auth_provider_settings.get_public_key()) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, f"Key for provider '{auth_provider_id}' unknown")
    try:
        payload = decode(token, key=key, algorithms=["RS256"], audience="oidc-test")
    except ExpiredSignatureError as err:
        logger.info(f"Expired signature: {err}")
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Expired signature (refresh not implemented yet)")
    except InvalidKeyError as err:
        logger.info(f"Invalid key: {err}")
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Invalid auth provider key")
    except Exception as err:
        logger.info("Cannot decode token, see below")
        logger.exception(err)
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Cannot decode token")
    if (user_id := payload.get('sub')) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Wrong token: 'sub' (user id) not found")
    try:
        user = await db.get_user(user_id)
    except UserNotInDB:
        logger.info(f"User {user_id} not found in DB, creating it (real apps can behave differently")
        user = await db.add_user(
            sub=payload['sub'],
            user_info=payload,
            oidc_provider=getattr(authlib_oauth, auth_provider_id),
            user_info_from_endpoint={}
        )
    return user

==> 0.1.38.zip: oidc-fastapi-test/src/oidc_test/auth_utils.py <==
from typing import Union, Annotated
from functools import wraps
import logging

from fastapi import HTTPException, Request, Depends, status
from fastapi.security import OAuth2PasswordBearer
from authlib.oauth2.rfc6749 import OAuth2Token
from authlib.integrations.starlette_client import OAuth, OAuthError, StarletteOAuth2App
from jwt import ExpiredSignatureError, InvalidKeyError, decode
from httpx import AsyncClient

# from authlib.oauth1.auth import OAuthToken
# from authlib.oauth2.auth import OAuth2Token

from .models import User
from .database import db, UserNotInDB
from .settings import settings, OIDCProvider

logger = logging.getLogger(__name__)

oidc_providers_settings: dict[str, OIDCProvider] = dict(
    [(provider.id, provider) for provider in settings.oidc.providers]
)

oauth2_scheme = OAuth2PasswordBearer(tokenUrl="token")


def fetch_token(name, request):
    breakpoint()
    ...
    # if name in oidc_providers:
    #    model = OAuth2Token
    # else:
    #    model = OAuthToken

    # token = model.find(name=name, user=request.user)
    # return token.to_token()


def update_token(*args, **kwargs):
    breakpoint()
    ...


authlib_oauth = OAuth(cache=None, fetch_token=fetch_token, update_token=update_token)


def init_providers():
    # Add oidc providers to authlib from the settings
    for id, provider in oidc_providers_settings.items():
        authlib_oauth.register(
            name=id,
            server_metadata_url=provider.openid_configuration,
            client_kwargs={
                "scope": "openid email offline_access profile",
            },
            client_id=provider.client_id,
            client_secret=provider.client_secret,
            api_base_url=provider.url,
            # For PKCE (not implemented yet):
            # code_challenge_method="S256",
            # fetch_token=fetch_token,
            # update_token=update_token,
            # client_id="some-client-id", # if enabled, authlib will also check that the access token belongs to this client id (audience)
        )


init_providers()


async def get_providers_info():
    # Get the public key:
    async with AsyncClient() as client:
        for provider_settings in oidc_providers_settings.values():
            if provider_settings.info_url:
                provider_info = await client.get(provider_settings.url)
                provider_settings.info = provider_info.json()


def get_oidc_provider_or_none(request: Request) -> StarletteOAuth2App | None:
    """Return the oidc_provider from a request object, from the session.
    It can be used in Depends()"""
    if (oidc_provider_id := request.session.get("oidc_provider_id")) is None:
        return
    try:
        return getattr(authlib_oauth, str(oidc_provider_id))
    except AttributeError:
        return


def get_oidc_provider(request: Request) -> StarletteOAuth2App:
    if (oidc_provider := get_oidc_provider_or_none(request)) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No such provider")
    else:
        return oidc_provider


async def get_current_user(request: Request) -> User:
    """Get the current user from a request object.
    Also validates the token expiration time.
    ... TODO: complete about refresh token
    """
    if (user_sub := request.session.get("user_sub")) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED)
    if (token := await db.get_token(request.session["token"])) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Token unknown")
    user = await db.get_user(user_sub)
    ## Check if the token is expired
    if token.is_expired():
        oidc_provider = get_oidc_provider(request=request)
        ## Ask a new refresh token from the provider
        logger.info(f"Token expired for user {user.name}")
        try:
            userinfo = await oidc_provider.fetch_access_token(
                refresh_token=token.get("refresh_token")
            )
        except OAuthError as err:
            logger.exception(err)
            # raise HTTPException(
            #    status.HTTP_401_UNAUTHORIZED, "Token expired, cannot refresh"
            # )

    return user


async def get_token(request: Request) -> OAuth2Token:
    """Return the token from a request object, from the session.
    It can be used in Depends()"""
    if (token := await db.get_token(request.session.get("token"))) is None:
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "No token")
    return token


async def get_current_user_or_none(request: Request) -> User | None:
    """Return the user from a request object, from the session.
    It can be used in Depends()"""
    try:
        return await get_current_user(request)
    except HTTPException:
        return None


def hasrole(required_roles: Union[str, list[str]] = []):
    """Decorator for RBAC permissions"""
    required_roles_set: set[str]
    if isinstance(required_roles, str):
        required_roles_set = set([required_roles])
    else:
        required_roles_set = set(required_roles)

    def decorator(func):
        @wraps(func)
        async def wrapper(request=None, *args, **kwargs):
            if request is None:
                raise HTTPException(
                    500,
                    "Functions decorated with hasrole must have a request:Request argument",
                )
            user: User = await get_current_user(request)
            if not any(required_roles_set.intersection(user.roles_as_set)):
                raise HTTPException(status.HTTP_401_UNAUTHORIZED)
            return await func(request, *args, **kwargs)

        return wrapper

    return decorator


def get_token_info(token: dict) -> dict:
    token_info = dict()
    for key in token:
        if key != "userinfo":
            token_info[key] = token[key]
    return token_info


async def get_user_from_token(
    token: Annotated[str, Depends(oauth2_scheme)],
    request: Request,
) -> User:
    if (auth_provider_id := request.headers.get("auth_provider")) is None:
        raise HTTPException(
            status.HTTP_401_UNAUTHORIZED,
            "Request headers must have a 'auth_provider' field",
        )
    if (
        auth_provider_settings := oidc_providers_settings.get(auth_provider_id)
    ) is None:
        raise HTTPException(
            status.HTTP_401_UNAUTHORIZED, f"Unknown auth provider '{auth_provider_id}'"
        )
    if (key := auth_provider_settings.get_public_key()) is None:
        raise HTTPException(
            status.HTTP_401_UNAUTHORIZED,
            f"Key for provider '{auth_provider_id}' unknown",
        )
    try:
        payload = decode(
            token,
            key=key,
            algorithms=["RS256"],
            audience="oidc-test",
            options={"verify_signature": not settings.insecure.skip_verify_signature},
        )
    except ExpiredSignatureError as err:
        logger.info(f"Expired signature: {err}")
        raise HTTPException(
            status.HTTP_401_UNAUTHORIZED,
            "Expired signature (refresh not implemented yet)",
        )
    except InvalidKeyError as err:
        logger.info(f"Invalid key: {err}")
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Invalid auth provider key")
    except Exception as err:
        logger.info("Cannot decode token, see below")
        logger.exception(err)
        raise HTTPException(status.HTTP_401_UNAUTHORIZED, "Cannot decode token")
    if (user_id := payload.get("sub")) is None:
        raise HTTPException(
            status.HTTP_401_UNAUTHORIZED, "Wrong token: 'sub' (user id) not found"
        )
    try:
        user = await db.get_user(user_id)
        user.access_token = payload
    except UserNotInDB:
        logger.info(
            f"User {user_id} not found in DB, creating it (real apps can behave differently"
        )
        user = await db.add_user(
            sub=payload["sub"],
            user_info=payload,
            oidc_provider=getattr(authlib_oauth, auth_provider_id),
            user_info_from_endpoint={},
            access_token=payload,
        )
    return user
